!+
! Module: rgssources_shared
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains routines shared between rgssources and rgssources_update.
! 
! $Id: rgssources_shared.f90,v 1.2 2000/08/11 13:08:45 pvideler Exp $
!- 
module rgssources_shared

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.
  use dal
!  use cal
  use errorHandling
  use oal
  use constants
  use rgssources_init
  use temp_rgslib
  use pointing_mod

  implicit none 

  integer(int16) :: formatNum
  character(20), dimension(0:1) :: tableNames = (/'SOURCES', 'SRCLIST'/)

  private

  public ucase				!***** put this in ssclib!
  public stringDump			!***** put this in ssclib!
  public no0s				!***** put this in ssclib!
  public calculateFOVcoords
  public invertFOVcoords
  public calculateConfusion
  public writeNewTable
  public writeGenericAttributes
  public writeObservationAttributes
  public writeExposureAttributes
  public strToInteger			!***** put this in ssclib!
  public intToStr			!***** put this in ssclib!

  public formatNum, tableNames

contains 

  !---------------------------------------------------------------------
  function ucase(arg) result(string)
  !---------------------------------------------------------------------

    implicit none 

    character(*), intent(in) :: arg
    character(len=len(arg)) :: string
    integer :: i, asciiIndex, aAsciiIndex, zAsciiIndex, bigAasciiIndex

    !-------------------------------------------------------------------
    ! Start:

    aAsciiIndex = ichar('a')
    zAsciiIndex = ichar('z')
    bigAasciiIndex = ichar('A')
    string = arg
    do i = 1, len(arg)
      asciiIndex = ichar(string(i:i))
      if (asciiIndex .ge. aAsciiIndex .and. asciiIndex .le. zAsciiIndex) then
        string(i:i) = char(bigAasciiIndex + asciiIndex - aAsciiIndex)
      end if
    end do

  end function ucase

  !---------------------------------------------------------------------
  subroutine stringDump(str)
  !---------------------------------------------------------------------

    implicit none 

    character(*), intent(in) :: str

    character(1) :: chr
    integer :: length, i, index

    !-------------------------------------------------------------------
    ! Start:

    length = len(str)

    if (length .le. 0) then
      print *, 'There are 0 characters in the string.'
  return

    else if (length .eq. 1) then
      print *, 'There is 1 character in the string.'

    else
      print *, 'There are', length, 'characters in the string.'
    end if
    print *, ''

    do i = 1, length
      chr = str(i:i)
      index = ichar(chr)
      if (index .le. 32 .or. index .ge. 127) then
        chr = ' '
      end if
      write(*, '(a3, z2)') chr//'  ', index
    end do

  end subroutine stringDump

  !---------------------------------------------------------------------
  function no0s(str)
  !---------------------------------------------------------------------

    implicit none 

    character(*), intent(in) :: str

    character(len=len(str)) :: no0s, tempstr
    character(1) :: chr
    integer :: length, i, index, lastChar

    !-------------------------------------------------------------------
    ! Start:

    length = len(str)
    if (length .le. 0) then
      no0s = str
  return
    end if

    lastChar = length
    do i = length, 1, -1
      chr = str(i:i)
      index = ichar(chr)
      if (index .ne. 0 .and. index .ne. ichar(' ')) then
        lastChar = i
    exit
      end if
      tempstr(i:i) = ' '
    end do
    tempstr(1:lastChar) = str(1:lastChar)

no0s = tempstr

  end function no0s

  !---------------------------------------------------------------------
  subroutine calculateFOVcoords(srcData, attitude)
  !---------------------------------------------------------------------

    implicit none 

    type(SpacecraftAttitudeType), intent(in) :: attitude ! typedef from oal

    type(sourceType), intent(inout) ::&
      srcData
    type(SkyPositionType) :: srcSkyPos
    real(double) :: offAxisAngles(2)

    !-------------------------------------------------------------------
    ! Start:

    srcSkyPos%ra  = srcData%raDeg  * piOn180
    srcSkyPos%dec = srcData%decDeg * piOn180

    offAxisAngles = raDecToOffAxis(attitude, srcSkyPos) ! (new in
    ! rgslib-1.14)
    ! routine from rgslib/pointing_mod. This in turn calls
    ! rgslib/dummycal_mod function RGSLIB_getFoverL, which needs a CCF.

    srcData%deltaDispArcmin = offAxisAngles(1) * 60.0 / piOn180
    srcData%deltaXdspArcmin = offAxisAngles(2) * 60.0 / piOn180

    call dispXdspToPolar(&
      srcData%deltaDispArcmin,&
      srcData%deltaXdspArcmin,&
      srcData%fovPhiDeg,&
      srcData%fovRdeg&
    )

  end subroutine calculateFOVcoords

  !---------------------------------------------------------------------
  subroutine invertFOVcoords(srcData, attitude)
  !---------------------------------------------------------------------

    implicit none 

    type(SpacecraftAttitudeType), intent(in) :: attitude ! typedef from oal

    type(sourceType), intent(inout) ::&
      srcData
    type(SkyPositionType) :: srcSkyPos
    real(double) :: offAxisAngles(2)

    !-------------------------------------------------------------------
    ! Start:

    offAxisAngles(1) = srcData%deltaDispArcmin * piOn180 / 60.0
    offAxisAngles(2) = srcData%deltaXdspArcmin * piOn180 / 60.0

    srcSkyPos = offAxisToRaDec(attitude, offAxisAngles) ! new in
    ! rgslib-1.14
    ! routine from rgslib/pointing_mod. This in turn calls
    ! rgslib/dummycal_mod function RGSLIB_getFoverL, which needs a CCF.

    srcData%raDeg  = srcSkyPos%ra  / piOn180
    srcData%decDeg = srcSkyPos%dec / piOn180

  end subroutine invertFOVcoords

  !---------------------------------------------------------------------
  subroutine calculateConfusion(srcData, primeSrc)
  !---------------------------------------------------------------------

    implicit none 

    type(sourceType), dimension(:), intent(inout) ::&
      srcData
    integer(int16), intent(in) ::&
      primeSrc

    integer(int16) ::&
      numSrcs
    integer(int16) ::&
      src
    real(single) :: delta_disp, delta_xdisp, iPrime
 
!****** these should come from CAL (CCF MiscData)
    real, parameter :: p1 = -0.00653
    real, parameter :: p2 =  1.01
    real, parameter :: a1 = 36.9
    real, parameter :: a2 = 593.4
    real, parameter :: b1 = 2.29
    real, parameter :: b2 = 23.3

    ! See package documentation for equation 
    ! offaxis angle converted to arcsec

character(10) :: primeStr
integer :: status

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Calculating RGS confusion value.')

    numSrcs = size(srcData)

    if (srcData(primeSrc)%rate .le. 0.0) then
call intToStr(int(primeSrc), 'left', primeStr, status)
!      write(errstr, '(a, i5, a)') 'The prime source (index ', primeSrc&
!      , ') has a RATE value of <= zero. Confusion values cannot be calculated.'
!      call warning('badPrimeRate')
write(errstr, '(a)') 'The prime source (index '//trim(primeStr)&
//') has a RATE value of <= zero. Confusion values cannot be calculated.'
call warning('badPrimeRate')

      srcData%confusion = 0.0

    else
      do src = 1, numSrcs
        delta_disp  = abs(srcData(src)%deltaDispArcmin&
        - srcData(primeSrc)%deltaDispArcmin)
        delta_xdisp = abs(srcData(src)%deltaXdspArcmin&
        - srcData(primeSrc)%deltaXdspArcmin)

        iPrime = p1 + p2 * srcData(src)%rate&
        * exp(-0.5*((delta_disp  * 60.0 + a1) / a2)**2)&
        * exp(-0.5*((delta_xdisp * 60.0 + b1) / b2)**2)
        srcData(src)%confusion = iPrime / srcData(primeSrc)%rate
      end do
    end if

  end subroutine calculateConfusion

  !---------------------------------------------------------------------
  subroutine writeNewTable(tab, srcData)
  !---------------------------------------------------------------------

    implicit none 

    type(TableT), intent(in) ::&
      tab
    type(sourceType), dimension(:), intent(in) ::&
      srcData

    integer(int16) ::&
      numSrcs, src
    type(ColumnT) ::&
      indexCol,&
      labelCol,&
      raCol,&
      decCol,&
      rateCol,&
      dispCol,&
      xdspCol,&
      phiCol,&
      rCol,&
      confnCol,&
      procCol,&
      excluCol,&
      fixedCol,&
      epicnCol,&
      flagCol
    integer(int16), dimension(:), pointer ::&
      indexColPtr
    integer(int32), dimension(:), pointer ::&
      epicnColPtr,&
      flagColPtr
    real(single), dimension(:), pointer ::&
      raColPtr,&
      decColPtr,&
      rateColPtr,&
      dispColPtr,&
      xdspColPtr,&
      phiColPtr,&
      rColPtr,&
      confnColPtr
    logical(bool), dimension(:), pointer ::&
      procColPtr,&
      excluColPtr,&
      fixedColPtr
    integer :: labelDims(1)

    !-------------------------------------------------------------------
    ! Start:

    numSrcs = size(srcData)

    labelDims(1) = int(labelsize)

    ! Set up columns:
    indexCol = addColumn(tab, 'INDEX',        INTEGER16)
    labelCol = addColumn(tab, 'LABEL',        STRING, dimensions=labelDims&
      , comment = 'This string is unique for each source.')
    raCol    = addColumn(tab, 'RA',           REAL32, 'degrees')
    decCol   = addColumn(tab, 'DEC',          REAL32, 'degrees')
    rateCol  = addColumn(tab, 'RATE',         REAL32, 'cts/s'&
      , comment = 'Taken from EPIC source list, or user-supplied.')
    dispCol  = addColumn(tab, 'DELTA_DISP',   REAL32, 'arcmin'&
      , comment = 'Offset from pointing: dispersion direction.')
    xdspCol  = addColumn(tab, 'DELTA_XDSP',   REAL32, 'arcmin'&
      , comment = 'Offset from pointing: cross-dispersion direction.')
    phiCol   = addColumn(tab, 'FOV_PHI',      REAL32, 'degrees'&
      , comment = 'Offset from pointing in polar coordinates: angle from &
      &the -ve dispersion axis towards the +ve cross-dispersion axis.')
    rCol     = addColumn(tab, 'FOV_R',        REAL32, 'degrees'&
      , comment = 'Offset from pointing in polar coordinates: radius from &
      &pointing.')
    confnCol = addColumn(tab, 'CONFUSION',    REAL32&
      , comment = 'Fraction giving a measure of the confusion between each &
      &source and the prime source.')
    procCol  = addColumn(tab, 'PROCESS',      BOOLEAN&
      , comment = 'True means a spectrum extraction region should be &
      &constructed for this source.')
    excluCol = addColumn(tab, 'BKG_EXCLUDE',  BOOLEAN&
      , comment = 'True means this source should be excluded from the &
      &background spectrum extraction region.')
    fixedCol = addColumn(tab, 'FIXED_ON_SKY', BOOLEAN&
      , comment = 'True for sources whose positions were initially entered &
      &as ra and dec.')
    epicnCol = addColumn(tab, 'EPIC_FILE',    INTEGER32&
      , comment = 'Key to identify the appropriate EPICFILEn and &
      &EPICBANDn attributes.')
    flagCol  = addColumn(tab, 'FLAG',         INTEGER32&
      , comment = 'Quality flag.')

    ! Open column data values:
    indexColPtr => int16Data( indexCol)
    raColPtr    => real32Data(   raCol)
    decColPtr   => real32Data(  decCol)
    rateColPtr  => real32Data( rateCol)
    dispColPtr  => real32Data( dispCol)
    xdspColPtr  => real32Data( xdspCol)
    phiColPtr   => real32Data(  phiCol)
    rColPtr     => real32Data(    rCol)
    confnColPtr => real32Data(confnCol)
    procColPtr  => boolData(   procCol)
    excluColPtr => boolData(  excluCol)
    fixedColPtr => boolData(  fixedCol)
    epicnColPtr => int32Data( epicnCol)
    flagColPtr  => int32Data(  flagCol)

    call setNullValue(epicnCol, myInt32null)
    if (.not. nullDefined(nullable(epicnCol))) then
call error('unableToSetNull', 'Could not set an int32 null value for &
&column EPIC_FILE.')
    end if

    ! write column values:
    do src = 1, numSrcs
      indexColPtr(src - 1) = srcData(src)%index
      raColPtr(src - 1)    = srcData(src)%raDeg
      decColPtr(src - 1)   = srcData(src)%decDeg
      rateColPtr(src - 1)  = srcData(src)%rate
      dispColPtr(src - 1)  = srcData(src)%deltaDispArcmin
      xdspColPtr(src - 1)  = srcData(src)%deltaXdspArcmin
      phiColPtr(src - 1)   = srcData(src)%fovPhiDeg
      rColPtr(src - 1)     = srcData(src)%fovRdeg
      confnColPtr(src - 1) = srcData(src)%confusion
      procColPtr(src - 1)  = srcData(src)%process
      excluColPtr(src - 1) = srcData(src)%bkgExclude
      fixedColPtr(src - 1) = srcData(src)%fixedOnSky
      flagColPtr(src - 1)  = srcData(src)%flag

      if (srcData(src)%epicFileAttrNum .eq. myInt32null) then
        call setToNull(epicnCol, src - 1)
      else
        epicnColPtr(src - 1) = srcData(src)%epicFileAttrNum
      end if

      call setStringCell(labelCol , src - 1, srcData(src)%label)
    end do

    call release(indexCol)
    call release(labelCol)
    call release(raCol)
    call release(decCol)
    call release(rateCol)
    call release(dispCol)
    call release(xdspCol)
    call release(phiCol)
    call release(rCol)
    call release(confnCol)
    call release(procCol)
    call release(excluCol)
    call release(fixedCol)
    call release(epicnCol)
    call release(flagCol)

  end subroutine writeNewTable

  !---------------------------------------------------------------------
  subroutine writeGenericAttributes(set, instrumentName, exposureNum, attitude)
  !---------------------------------------------------------------------

    implicit none 

    type(DataSetT), intent(in) :: set
    character(*), intent(in) :: instrumentName
    integer(int32), intent(in) :: exposureNum
    type(SpacecraftAttitudeType), intent(in) :: attitude ! typedef from oal

    !-------------------------------------------------------------------
    ! Start:

    call setAttribute(set, 'TELESCOP', 'XMM')
    call setAttribute(set, 'INSTRUME', trim(instrumentName))
    call setAttribute(set, 'EXP_ID', exposureNum)
    call setAttribute(set, 'FILTER', 'UNKNOWN'&
    , comment='RGS is filterless.')
    call setAttribute(set, 'ORIGIN', 'Leicester/SSC'&
    , comment="Origin of FITS file")
    call setAttribute(set, 'RA_PNT',  attitude%ra   / piOn180)
    call setAttribute(set, 'DEC_PNT', attitude%dec  / piOn180)
    call setAttribute(set, 'PA_PNT',  attitude%apos / piOn180)
    call setAttribute(set, 'RADECSYS', 'FK5')
    call setAttribute(set, 'EQUINOX', '2000.0')
!    call setAttribute(set, 'EXPOSURE',)
!    call setAttribute(set, 'DATE-OBS',)
!    call setAttribute(set, 'DATE-END',)
!    call setAttribute(set, 'OCB',) ! special RGS header kwd: on-chip binning

  end subroutine writeGenericAttributes

  !---------------------------------------------------------------------
  subroutine writeObservationAttributes(set)
  !---------------------------------------------------------------------

    use oal

    implicit none 

    type(DataSetT), intent(in) :: set

    type(ProposalInfoType) :: proposalInfo
    type(OdfInfoType)      :: odfInfo

    !-------------------------------------------------------------------
    ! Start:

    call OAL_odfInfo(odfInfo) 
    call OAL_proposalInfo(proposalInfo)

    if (odfInfo%isSlew) then
      call setAttribute(set, "OBS_MODE", "SLEW"&
      , comment="Observation mode (pointing or slew)")
    else
      call setAttribute(set, "OBS_MODE", "POINTING"&
      , comment="Observation mode (pointing or slew)")
    endif
    call setAttribute(set, 'OBS_ID', trim(odfInfo%observationId)&
    , comment="Observation identifier")
    call setAttribute(set, "OBJECT", trim(proposalInfo%targetName)&
    , comment="Name of observed object")
    call setAttribute(set, "OBSERVER", trim(proposalInfo%GoName)&
    , comment="Name of observer")

  end subroutine writeObservationAttributes

  !---------------------------------------------------------------------
  subroutine writeExposureAttributes(set, instrumentId, exposureId)
  !---------------------------------------------------------------------

    use oal

    implicit none 

    type(DataSetT), intent(in) :: set
    integer(int32), intent(in) :: instrumentId
    character(4),   intent(in) :: exposureId

    type(ExposureInfoType) :: exposureInfo
    integer(int32)         :: dataMode

    !-------------------------------------------------------------------
    ! Start:

    call OAL_setState(instrument=instrumentId, exposureId=exposureId)
    call OAL_getState(dataMode=dataMode)
    call OAL_exposureInfo(exposureInfo) 

    call setAttribute(set, 'DATAMODE', dataModeId(dataMode)&
    , comment='Instrument mode.')
    call setAttribute(set, 'EXP_ID', exposureInfo%number&
    , comment="Exposure identifier")

  end subroutine writeExposureAttributes

  !---------------------------------------------------------------------
  subroutine writeExposureAttributes_old(set, instrumentId, expId)
  !---------------------------------------------------------------------

    use oal

    implicit none 

    type(DataSetT), intent(in) :: set
    integer(int32), intent(in) :: instrumentId, expId

    type(ExposureInfoType) :: exposureInfo
    integer(int32)         :: dataMode

    !-------------------------------------------------------------------
    ! Start:

    call OAL_setState(instrument=instrumentId, exposureNr=expId)
    call OAL_getState(dataMode=dataMode)
    call OAL_exposureInfo(exposureInfo) 

    call setAttribute(set, 'DATAMODE', dataModeId(dataMode)&
    , comment='Instrument mode.')
    call setAttribute(set, 'EXP_ID', exposureInfo%number&
    , comment="Exposure identifier")

  end subroutine writeExposureAttributes_old

  !---------------------------------------------------------------------
  subroutine strToInteger(str, int, status)
  !---------------------------------------------------------------------

    implicit none 

    character(*), intent(in) :: str
    integer, intent(out) :: int, status

    integer :: i, length, nonZeroLength, mult

    !-------------------------------------------------------------------
    ! Start:

    status = 0 ! default
    int    = 0 ! default

    length = len(str)
    if (length .le. 0) then
      status = 1
  return
    end if

    do i = 1, length
      if (iachar(str(i:i)) .lt. iachar('0')&
      .or. iachar(str(i:i)) .gt. iachar('9')) then
        status = 2
  return
      end if
    end do

    nonZeroLength = 0
    do i = 1, length
      if (iachar(str(i:i)) .ne. iachar('0')) then
        nonZeroLength = length + 1 - i
      end if
    end do

    if (nonZeroLength .eq. 0) then
      int = 0
  return
    end if

    if (nonZeroLength .gt. floor(log10(real(huge(int))))) then
    ! floor because eg huge(int16) is 32768, thus ceiling(log10(huge(int16)))
    ! is 5, allowing a str value of, for example, '32769', which would cause
    ! an overflow.
      status = 3
  return
    end if

    int = iachar(str(length:length)) - iachar('0')
    mult = 1
    do i = length - 1, 1 + length - nonZeroLength, -1
      mult = mult * 10
      int = int + mult * iachar(str(i:i)) - iachar('0')
    end do

  end subroutine strToInteger

  !---------------------------------------------------------------------
  subroutine intToStr(int, justify, str, status)
  !---------------------------------------------------------------------

    implicit none

    integer,      intent(in)  :: int
    character(*), intent(in)  :: justify
    character(*), intent(out) :: str
    integer,      intent(out) :: status

    integer :: numCharRequired

    !-------------------------------------------------------------------
    ! Start

    status = 0 ! default
    str = '' ! default

    numCharRequired = 1 + log10(real(int))
    if (int .lt. 0) numCharRequired = numCharRequired + 1
    if (numCharRequired .gt. len(str)) then
      status = -1
  return
    end if

    write(str, *) int
    if (justify .eq. 'left') then
      str = adjustl(str)
    else
      str = adjustr(str)
    end if

  end subroutine intToStr

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template
end module rgssources_shared






