!+
! Module: temp_rgslib
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: Contains coordinate transform routines that won't appear in rgslib until version 1.14 or later.
! 
! $Id: temp_rgslib.f90,v 1.2 2000/08/11 13:08:45 pvideler Exp $
!- 
module temp_rgslib

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.
  use dal
!  use cal
  use errorHandling
  use oal		! -> SpacecraftAttitudeType
  use constants
  use pointing_mod	! -> SkyPositionType

  implicit none 

  private 
  public dispXdspToPolar, raDecToOffAxis, offAxisToRaDec

!  type skyPositionType
!    real(double) :: ra, dec  ! ra dec of position in sky.  (radians epoch 2000)
!  end type skyPositionType

contains 

  !---------------------------------------------------------------------
  subroutine dispXdspToPolar(deltaDispArcmin, deltaXdspArcmin&
  , fovPhiDeg, fovRdeg)
  !---------------------------------------------------------------------

    use dummycal_mod

    implicit none 

    real(single), intent(in)  :: deltaDispArcmin, deltaXdspArcmin
    real(single), intent(out) :: fovPhiDeg, fovRdeg
    real(single) :: weightedDD
  
    !-------------------------------------------------------------------
    ! Start:

    weightedDD = -deltaDispArcmin / RGSLIB_getFoverL() ! The FoverL thing
    ! is in rgslib/dummycal_mod.

    if (deltaXdspArcmin .eq. 0.0 .and. weightedDD .eq. 0.0) then
      fovPhiDeg = 0.0
    else
      fovPhiDeg = atan2(deltaXdspArcmin, weightedDD) / piOn180
    end if
    fovRdeg = sqrt(weightedDD * weightedDD&
    + deltaXdspArcmin * deltaXdspArcmin) / 60.0

  end subroutine dispXdspToPolar

  !---------------------------------------------------------------------
  function raDecToOffAxis(attitude, srcSkyPos) result(offAxisAngles)
  !---------------------------------------------------------------------
  ! This function is supplied with the spacecraft attitude (point direction
  ! + position angle) and the ra and dec of the source (= srcSkyPos). Its
  ! job is (i) to convert the attitude into an orthnormal basis, the three
  ! basis vectors being expressed in sky-basis coordinates; (ii) convert
  ! the source ra and dec into a vector of direction cosines (= srcVector_sky),
  ! again expressed in sky-basis coordinates; (iii) change the basis of
  ! this vector to the attitude basis (done by premultiplying the vector
  ! by a matrix whose rows consist of the attitude vectors, expressed in
  ! sky-basis coordinates); (iv) finally, to convert the new vector
  ! (srcVector_att) into two offaxis angles.
  ! 
  ! The procedure is straightforward in the case that there is a sizeable
  ! angle between srcVector and the negative x axis of the
  ! attitude basis. It is however inaccurate in the contrary case: it can
  ! be shown that the calculation is becomes proportional to 1.0 minus the
  ! cos of this angle. Hence an approximate algorithm has been substituded
  ! for this latter case, in which the 1 - cos has been replaced by the
  ! first three terms of its Taylor expansion.
  !
  ! The function can be used in place of the (slightly less accurate under
  ! some circumstances) offAxisAngle_sc().

  ! NOTE: CAL has to be initialised with CAL_setState() before this is called. 

    use vector_mod
    use dummycal_mod

    implicit none

    type(SpacecraftAttitudeType), intent(in) :: attitude 
    type(SkyPositionType)       , intent(in) :: srcSkyPos
    real(double), dimension(2)               :: offAxisAngles

    type(SpacecraftAttitudeType) ::&
      dummyAttitude 
    real(double), dimension(3,3) ::&
      attMatrix_sky,&
      attMatrix_A2,&
      dummyAttMatrix_sky
    real(double), dimension(3) ::&
      srcVector_sky,&
      srcVector_A2,&
      srcVector_att
    real(double) ::&
      deltaRa,&
      deltaDec,&
      fourthTermOnFirst,&
      cosSrcDec,&
      oneMinusCosDeltaRa,&
      alignment
    logical ::&
      matrixIsAsExpected,&
      deltaRaIsSmall

    !-------------------------------------------------------------------
    ! Start

    dummyAttitude%ra   = 0.0
    dummyAttitude%dec  = 0.0
    dummyAttitude%apos = attitude%apos
    call OAL_toAttitudeMatrix(dummyAttitude, dummyAttMatrix_sky)
    dummyAttMatrix_sky = transpose(dummyAttMatrix_sky) ! because it comes out in
    ! c-lish format, with attitudeMatrix(:,i) equal to the ith vector of
    ! the orthonormal attitude basis.

    alignment = sum(dummyAttMatrix_sky(1, :) * (/-1.0, 0.0, 0.0/))
    if (-log10(1.0d0 - abs(alignment))&
    .gt. 2 + precision(dummyAttMatrix_sky(1, 1))) then ! the 2 is just to
    ! provide a bit of leeway
      matrixIsAsExpected = .true.
    else
      matrixIsAsExpected = .false.

      call warning('badAttMatrix', 'rgslib: the x-axis of the attitude &
      &matrix does not point in the opposite direction to the spacecraft. &
      &Using fallback algorithm to calculate off-axis angles.')
    end if

    deltaRa = srcSkyPos%ra - attitude%ra 
    fourthTermOnFirst = deltaRa * deltaRa * deltaRa * deltaRa * deltaRa&
    * deltaRa / 20160.0d0
    if (-log10(fourthTermOnFirst) .gt. precision(deltaRa)) then
      deltaRaIsSmall = .true.
    else
      deltaRaIsSmall = .false.
    end if

    if (matrixIsAsExpected .and. deltaRaIsSmall) then ! use the approximate
    ! algorithm that involves an expansion of 1 - cos(deltaRa).
    !
    ! The change of basis is carried out in three steps, or via two
    ! intermediate bases and one final one. The first intermediate basis
    ! is one which has the same ra as the attitude but zero dec; the apos
    ! is such that its z vector (remember that the x vector of the attitude
    ! frame is assumed to point in the opposite direction to the spacecraft)
    ! (this is checked for aposMatrix above) points in the negative z
    ! direction in the sky. In other words, its components in the sky basis
    ! are as follows:
    !
    !	         (-cos(attitude%ra)   -sin(attitude%ra)      0) <- x axis
    !	A1_sky = ( sin(attitude%ra)   -cos(attitude%ra)      0) <- y axis
    !	         (        0                    0            -1) <- z axis
    !
    ! Transformation to this basis is only implied however since the
    ! components of the source location vector are not explicitly calculated
    ! in this basis. The second intermediate basis has the same ra and dec
    ! as the attitude basis but the apos is the same as the previous. The
    ! coefficients of the source location vector are calculated using
    ! analytic expressions for the components in the first intermediate
    ! basis of both the source location vector and the second intermediate
    ! basis.
    !
    ! The components of the 2nd IB in terms of the 1st are:
    !
    !	        ( cos(attitude%dec)   0      sin(attitude%dec)) <- x axis
    !	A2_A1 = (        0            1             0         ) <- y axis
    !	        (-sin(attitude%dec)   0      cos(attitude%dec)) <- z axis
    !
    ! The components of the source direction vector in the same basis are:
    !
    !	       (-cos(srcSkyPos%dec) cos(deltaRa) )
    !	s_A1 = (-cos(srcSkyPos%dec) sin(deltaRa) )
    !	       (-sin(srcSkyPos%dec)              )
    !
    ! where deltaRa = srcSkyPos%ra - attitude%ra. The analytical expressions
    ! for s_A2 = A2_A1 * s_A1 contain the expression (1 - cos(deltaRa)).
    ! This expression is replaced by the first 3 terms of its Taylor
    ! expansion, which is more precise when the 4th/1st term is less than
    ! the precision of the variable type. The expressions for s_A2 are
    ! then evaluated, giving the components of the source direction in
    ! terms of the 2nd intermediate basis.
    !
    ! The quickest way to perform the second and final part of the algorithm,
    ! to give the expression of the source direction in the full attitude
    ! basis, is to start by obtaining the attitude matrix from a dummy
    ! attitude with zero ra and dec but the same apos as the real attitude.
    ! If all its components are negated, this basis becomes the true
    ! attitude expressed in terms of the second intermediate basis. To see
    ! this, let us view the changes of basis in terms of the equivalent
    ! inversions/rotations. Let the same inversions/rotations be applied to
    ! both the full attitude frame and the source direction, both in sky
    ! coordinates. Expressing the source direction in terms of intermediate
    ! basis 1 is equivalent to an inversion (negation of all coordinates)
    ! and a rotation by -attitude%ra. The shift to the second intermediate
    ! basis is equivalent to a rotation by -attitude%dec. The coordinate
    ! values of the source direction vector are now the same as we calculated
    ! for s_A2: for a source close to the s/c pointing, the x coordinate
    ! should be close to -1 and the magnitudes of the y and z components
    ! should be small. The rotated attitude frame is the same as we would
    ! get by calculating the attitude matrix aposMatrix for the dummy
    ! attitude with ra = dec = 0, then negating all values. Note that the
    ! negation operation commutes with any rotation and hence can be applied
    ! at the end rather than the beginning.

      cosSrcDec = cos(srcSkyPos%dec)
      deltaDec = srcSkyPos%dec - attitude%dec 

      ! this is the approximation:
      oneMinusCosDeltaRa = deltaRa * deltaRa / 2.0d0 * (1.0d0 - deltaRa&
      * deltaRa / 12.0d0 * (1.0d0 - deltaRa * deltaRa / 30.0d0))

      srcVector_A2(1) = cos(attitude%dec) * cosSrcDec * oneMinusCosDeltaRa&
      - cos(deltaDec)
      srcVector_A2(2) = -cosSrcDec * sin(deltaRa)
      srcVector_A2(3) = -sin(attitude%dec) * cosSrcDec * oneMinusCosDeltaRa&
      - sin(deltaDec)

      ! Correct for apos (note negation of the dummy matrix, as described
      ! above):
      attMatrix_A2 = -dummyAttMatrix_sky
      srcVector_att = matmul(attMatrix_A2, srcVector_A2)

    else ! use the brute-force algorithm:
      call OAL_toAttitudeMatrix(attitude, attMatrix_sky)
      attMatrix_sky = transpose(attMatrix_sky) ! because it comes out in
      ! c-lish format, with attMatrix_sky(:,i) equal to the ith vector of
      ! the orthonormal attitude basis.

      srcVector_sky = unitCartesian(srcSkyPos%ra, srcSkyPos%dec)
      srcVector_att = matmul(attMatrix_sky, srcVector_sky)
    end if

if (srcVector_att(1) .ge. 0.0) call error('srcTooFarOffAxis', 'rgslib: the &
&source is greater than 90 degrees away from the pointing!')

    offAxisAngles(1)  = atan2(-srcVector_att(3), -srcVector_att(1))&
    * RGSLIB_getFoverL()
    ! dispersion direction

    offAxisAngles(2)  = -atan2(-srcVector_att(2), -srcVector_att(1))
    ! cross-dispersion direction

  end function raDecToOffAxis

  !---------------------------------------------------------------------
  function offAxisToRaDec(attitude, offAxisAngles) result(srcSkyPos)
  !---------------------------------------------------------------------
  ! NOTE: CAL has to be initialised with CAL_setState() before this is called. 

    use dummycal_mod

    implicit none

    type(SpacecraftAttitudeType), intent(in) :: attitude 
    real(double), dimension(2),   intent(in) :: offAxisAngles
    type(SkyPositionType)                    :: srcSkyPos

    real(double) ::&
      skyMatrix_att(3, 3),&
      semiRadius,&
      tan1,&
      tan2
    real(double), dimension(3) ::&
      srcVector_sky,&
      srcVector_att

    !-------------------------------------------------------------------
    ! Start

    ! First convert off-axis angles to off-axis vector components. Have to
    ! do this a little carefully because the off-axis angles as they are
    ! presently defined have pincushion symmetry rather than spherical.

    tan1 = tan(offAxisAngles(1) / RGSLIB_getFoverL())
    tan2 = tan(-offAxisAngles(2))
    srcVector_att(1) = -1.0d0 / sqrt(1.0d0 + tan1 * tan1 + tan2 * tan2)
    srcVector_att(2) = srcVector_att(1) * tan2
    srcVector_att(3) = srcVector_att(1) * tan1

    ! Now get the transpose of the attitude matrix, which is the same as its
    ! inverse, which is the same as the sky basis in attitude coordinates.
    ! Note that the call returns the matrix in c format, so it is already
    ! transposed - ie, the vector skyMatrix_att(i,:) should be the ith vector
    ! of the orthonormal sky basis (with components expressed in the attitude
    ! basis).

    call OAL_toAttitudeMatrix(attitude, skyMatrix_att)

    ! Now perform the matrix multiplication to convert the vector from
    ! attitude basis to sky basis, and extract the ra and dec.

    srcVector_sky = matmul(skyMatrix_att, srcVector_att)
    semiRadius = sqrt(srcVector_sky(1) * srcVector_sky(1)&
    + srcVector_sky(2) * srcVector_sky(2))
    if (semiRadius .eq. 0.0) then
      srcSkyPos%ra = 0.0
      srcSkyPos%dec = PiOverTwo * sign(1.0d0, srcVector_sky(3))
    else
      srcSkyPos%ra = atan2(srcVector_sky(2), srcVector_sky(1))
      if (srcSkyPos%ra .lt. 0.0) srcSkyPos%ra = srcSkyPos%ra + TwoPi 
      srcSkyPos%dec = atan(srcVector_sky(3) / semiRadius)
    end if

  end function offAxisToRaDec
end module temp_rgslib






